import numpy as np

from . import Message
from .std_msgs import Header


class LaserScan(Message):
    def __init__(self, ranges=[], intensitites=[]):
        self.header = Header()
        self.angle_min = None
        self.angle_max = None
        self.angle_increment = None
        self.time_increment = None
        self.scan_time = None
        self.range_min = None
        self.range_max = None
        self.ranges = np.array(ranges)
        # Change NaNs to range_max
        if np.isnan(self.ranges).any():
            self.ranges[np.isnan(self.ranges)] = self.range_max
        self.intensities = np.array(intensitites)

    def __str__(self):
        msg = "LaserScan Message\n"
        msg += str(self.header)
        msg += "Angle Min:    " + str(self.angle_min) + "\n"
        msg += "Angle Max:    " + str(self.angle_max) + "\n"
        msg += "Angle Inc:    " + str(self.angle_increment) + "\n"
        msg += "Time Inc:     " + str(self.time_increment) + "\n"
        msg += "Scan Time:    " + str(self.scan_time) + "\n"
        msg += "Range Min:    " + str(self.range_min) + "\n"
        msg += "Range Max:    " + str(self.range_max) + "\n"
        msg += "Ranges:       " + str(self.ranges) + "\n"
        msg += "Intensities:  " + str(self.intensities) + "\n"
        return msg

    def to_json(self):
        return {
            "header": self.header.to_json(),
            "angle_min": self.angle_min,
            "angle_max": self.angle_max,
            "angle_increment": self.angle_increment,
            "time_increment": self.time_increment,
            "scan_time": self.scan_time,
            "range_min": self.range_min,
            "range_max": self.range_max,
            "ranges": self.ranges.tolist(),
            "intensities": self.intensities.tolist(),
        }

    def from_json(self, msg):
        self.header.from_json(msg["header"])
        self.angle_min = msg["angle_min"]
        self.angle_max = msg["angle_max"]
        self.angle_increment = msg["angle_increment"]
        self.time_increment = msg["time_increment"]
        self.scan_time = msg["scan_time"]
        self.range_min = msg["range_min"]
        self.range_max = msg["range_max"]
        self.ranges = np.array(msg["ranges"])
        self.intensities = np.array(msg["intensities"])



import time
from datetime import datetime


def sleep(duration):
    if duration < 0.0:
        return
    else:
        time.sleep(duration)


class Rate:
    """
    Convenience class for sleeping in a loop at a specified rate
    """

    def __init__(self, frequency_hz):
        """
        Constructor.
        @param frequency_hz: Frequency rate in Hz to determine sleeping
        @type  frequency_hz: float
        """
        self.last_time = float(datetime.timestamp(datetime.utcnow()))
        self.sleep_dur = 1.0 / frequency_hz

    def _remaining(self, curr_time):
        """
        Calculate the time remaining for rate to sleep.
        @param curr_time: current time
        @type  curr_time: L{Time}
        @return: time remaining
        @rtype: L{Time}
        """
        # detect time jumping backwards
        if self.last_time > curr_time:
            self.last_time = curr_time

        # calculate remaining time
        elapsed = curr_time - self.last_time
        return self.sleep_dur - elapsed

    def remaining(self):
        """
        Return the time remaining for rate to sleep.
        @return: time remaining
        @rtype: L{Time}
        """
        curr_time = float(datetime.timestamp(datetime.utcnow()))
        return self._remaining(curr_time)

    def sleep(self):
        """
        Attempt sleep at the specified rate. sleep() takes into
        account the time elapsed since the last successful
        sleep().
        """
        curr_time = float(datetime.timestamp(datetime.utcnow()))
        sleep(self._remaining(curr_time))
        self.last_time = self.last_time + self.sleep_dur

        # detect time jumping forwards, as well as loops that are
        # inherently too slow
        if curr_time - self.last_time > self.sleep_dur * 2:
            self.last_time = curr_time


if __name__ == "__main__":
    import random

    r = Rate(10.0)
    curr_time = float(datetime.timestamp(datetime.utcnow()))
    while True:
        diff = 1.0 / (float(datetime.timestamp(datetime.utcnow())) - curr_time)
        curr_time = float(datetime.timestamp(datetime.utcnow()))
        x = random.random() / 10.0
        print(diff, x)
        # Random time-lenght long computation
        time.sleep(x)
        r.sleep()



import threading
from datetime import datetime


class Timer:
    def __init__(self, interval, callback):
        self.interval = interval
        self.callback = callback
        self.thread = threading.Timer(self.interval, self._calback_wrapper)
        self.thread.start()

    def _calback_wrapper(self):
        last_time = float(datetime.timestamp(datetime.utcnow()))
        self.callback()
        elapsed = float(datetime.timestamp(datetime.utcnow())) - last_time
        sleep_for = self.interval - elapsed
        self.thread = threading.Timer(sleep_for, self._calback_wrapper)
        self.thread.start()

    def stop(self):
        self.thread.cancel()


def example_callback():
    curr_time = float(datetime.timestamp(datetime.utcnow()))
    print("Callback called at: ", curr_time)


if __name__ == "__main__":
    r = Timer(1.0, example_callback)
    try:
        while True:
            pass
    except KeyboardInterrupt:
        r.cancel()
        print("Timer stopped")



import datetime
import json
from pathlib import Path

from zeroros.messages import Message


class DataLogger:
    def __init__(self, log_dir: str = "logs"):
        self.log_dir = log_dir
        # create filename with date and time
        self.filename = datetime.datetime.now().strftime("%Y%m%d_%H%M%S") + "_log.json"
        self.log_file = Path(self.log_dir) / self.filename
        self.log_file.parent.mkdir(parents=True, exist_ok=True)
        # Open the log file
        self.file = open(self.log_file, "w")

    def log(self, msg: type[Message]):
        # Write message to file
        self.file.write(
            '{"class": "'
            + str(type(msg).__name__)
            + '", "message": '
            + json.dumps(msg.to_json())
            + "}\n"
        )
        self.file.flush()



import threading

import zmq


class MessageBroker:
    def __init__(self, ip="127.0.0.1", xsub_port=5555, xpub_port=5556):
        # gather list of topics from pub/subs
        self.topics = {}
        self.context = None
        self.ip = ip
        self.xsub_port = xsub_port
        self.xpub_port = xpub_port
        # run broker in seperate thread
        self.broker_thread = threading.Thread(target=self.config_broker)
        self.broker_thread.start()

    def config_broker(self):
        """connects, binds, and configure sockets"""
        # https://zguide.zeromq.org/docs/chapter2/#The-Dynamic-Discovery-Problem
        try:
            self.context = zmq.Context()
            frontend = self.context.socket(zmq.XSUB)
            frontend.bind("tcp://" + str(self.ip) + ":" + str(self.xsub_port))
            backend = self.context.socket(zmq.XPUB)
            backend.bind("tcp://" + str(self.ip) + ":" + str(self.xpub_port))

            # built-in pub/sub fowarder
            zmq.proxy(frontend, backend)
        except zmq.error.ContextTerminated:
            print("Context terminated")
        except Exception as e:
            print("Error: ", e)
        finally:
            frontend.close()
            backend.close()

    def stop(self):
        self.context.term()


if __name__ == "__main__":
    MessageBroker()



import json
import time

import zmq

from zeroros.messages import Header, Message
from zeroros.topic import validate_topic


class Publisher:
    def __init__(
        self,
        topic: str,
        message_class: type[Message],
        ip: str = "127.0.0.1",
        port: int = 5555,
        verbose: bool = False,
    ):
        if verbose:
            print("Creating publisher for topic: ", topic)
        self.topic = validate_topic(topic)
        self.message_class = message_class
        self.ip = ip
        self.port = port
        self.context = zmq.Context()
        self.sock = self.context.socket(zmq.PUB)
        try:
            self.sock.connect("tcp://" + str(self.ip) + ":" + str(self.port))
        except Exception as e:
            print("Exception: ", e)
            raise e

    def publish(self, message):
        # Chech message is of the correct type
        if not isinstance(message, self.message_class):
            raise TypeError(f"Message must be of type {self.message_class.__name__}")
        message_str = json.dumps(message.to_json())
        topic_message_str = f"{self.topic} {message_str}"
        message_bytes = topic_message_str.encode("utf-8")
        self.sock.send(message_bytes)


if __name__ == "__main__":
    pub = Publisher("/header", Header)
    msg = Header()
    while True:
        msg.stamp = time.time()
        pub.publish(msg)
        msg.seq += 1
        time.sleep(0.5)
